# Auto generated from ontology_model.yaml by pythongen.py version: 0.9.0
# Generation date: 2022-06-30T15:03:19
# Schema: kgcl_schema-ontology-model
#
# id: https://w3id.org/kgcl/ontology
# description: A basic bare-bones model of an ontology or ontology-like structure. The purpose is not to provide a
#              complete model, rather just sufficient structure for domain and range constraints in the ocl model
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
import re
from jsonasobj2 import JsonObj, as_dict
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from linkml_runtime.linkml_model.meta import (
    EnumDefinition,
    PermissibleValue,
    PvFormulaOptions,
)

from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.metamodelcore import empty_list, empty_dict, bnode
from linkml_runtime.utils.yamlutils import (
    YAMLRoot,
    extended_str,
    extended_float,
    extended_int,
)
from linkml_runtime.utils.dataclass_extensions_376 import (
    dataclasses_init_fn_with_kwargs,
)
from linkml_runtime.utils.formatutils import camelcase, underscore, sfx
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from rdflib import Namespace, URIRef
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.linkml_model.types import String

metamodel_version = "1.7.0"
version = "0.0.1"

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
BASICS = CurieNamespace("basics", "https://w3id.org/kgcl/basics/")
DCTERMS = CurieNamespace("dcterms", "http://purl.org/dc/terms/")
LINKML = CurieNamespace("linkml", "https://w3id.org/linkml/")
OIO = CurieNamespace("oio", "http://www.geneontology.org/formats/oboInOwl#")
OM = CurieNamespace("om", "http://w3id.org/kgcl/om/")
OWL = CurieNamespace("owl", "http://www.w3.org/2002/07/owl#")
RDF = CurieNamespace("rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#")
XML = CurieNamespace("xml", "http://example.org/UNKNOWN/xml/")
DEFAULT_ = OM


# Types
class LanguageTag(str):
    type_class_uri = XML.lang
    type_class_curie = "xml:lang"
    type_name = "language tag"
    type_model_uri = OM.LanguageTag


# Class references
class NodeId(extended_str):
    pass


class ClassNodeId(NodeId):
    pass


class InstanceNodeId(NodeId):
    pass


class OntologyElement(YAMLRoot):
    """
    Any component of an ontology or knowledge graph
    """

    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OM.OntologyElement
    class_class_curie: ClassVar[str] = "om:OntologyElement"
    class_name: ClassVar[str] = "ontology element"
    class_model_uri: ClassVar[URIRef] = OM.OntologyElement


@dataclass
class PropertyValue(OntologyElement):
    """
    a property-value pair
    """

    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OM.PropertyValue
    class_class_curie: ClassVar[str] = "om:PropertyValue"
    class_name: ClassVar[str] = "property value"
    class_model_uri: ClassVar[URIRef] = OM.PropertyValue

    property: Optional[Union[str, NodeId]] = None
    filler: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.property is not None and not isinstance(self.property, NodeId):
            self.property = NodeId(self.property)

        if self.filler is not None and not isinstance(self.filler, str):
            self.filler = str(self.filler)

        super().__post_init__(**kwargs)


@dataclass
class Annotation(PropertyValue):
    """
    owl annotations. Not to be confused with annotations sensu GO
    """

    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OM.Annotation
    class_class_curie: ClassVar[str] = "om:Annotation"
    class_name: ClassVar[str] = "annotation"
    class_model_uri: ClassVar[URIRef] = OM.Annotation

    property: Optional[Union[str, NodeId]] = None
    filler: Optional[str] = None
    annotation_set: Optional[Union[dict, "Annotation"]] = None
    property_type: Optional[str] = None
    filler_type: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.property is not None and not isinstance(self.property, NodeId):
            self.property = NodeId(self.property)

        if self.filler is not None and not isinstance(self.filler, str):
            self.filler = str(self.filler)

        if self.annotation_set is not None and not isinstance(
            self.annotation_set, Annotation
        ):
            self.annotation_set = Annotation(**as_dict(self.annotation_set))

        if self.property_type is not None and not isinstance(self.property_type, str):
            self.property_type = str(self.property_type)

        if self.filler_type is not None and not isinstance(self.filler_type, str):
            self.filler_type = str(self.filler_type)

        super().__post_init__(**kwargs)


@dataclass
class Node(OntologyElement):
    """
    Any named entity in an ontology. May be a class, individual, property
    """

    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OM.Node
    class_class_curie: ClassVar[str] = "om:Node"
    class_name: ClassVar[str] = "node"
    class_model_uri: ClassVar[URIRef] = OM.Node

    id: Union[str, NodeId] = None
    name: Optional[str] = None
    annotation_set: Optional[Union[dict, Annotation]] = None
    owl_type: Optional[Union[str, "OwlTypeEnum"]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.id):
            self.MissingRequiredField("id")
        if not isinstance(self.id, NodeId):
            self.id = NodeId(self.id)

        if self.name is not None and not isinstance(self.name, str):
            self.name = str(self.name)

        if self.annotation_set is not None and not isinstance(
            self.annotation_set, Annotation
        ):
            self.annotation_set = Annotation(**as_dict(self.annotation_set))

        if self.owl_type is not None and not isinstance(self.owl_type, OwlTypeEnum):
            self.owl_type = OwlTypeEnum(self.owl_type)

        super().__post_init__(**kwargs)


@dataclass
class ClassNode(Node):
    """
    A node that is a class
    """

    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OWL.Class
    class_class_curie: ClassVar[str] = "owl:Class"
    class_name: ClassVar[str] = "class node"
    class_model_uri: ClassVar[URIRef] = OM.ClassNode

    id: Union[str, ClassNodeId] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.id):
            self.MissingRequiredField("id")
        if not isinstance(self.id, ClassNodeId):
            self.id = ClassNodeId(self.id)

        super().__post_init__(**kwargs)


@dataclass
class InstanceNode(Node):
    """
    A node that is an individual
    """

    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OWL.NamedIndividual
    class_class_curie: ClassVar[str] = "owl:NamedIndividual"
    class_name: ClassVar[str] = "instance node"
    class_model_uri: ClassVar[URIRef] = OM.InstanceNode

    id: Union[str, InstanceNodeId] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.id):
            self.MissingRequiredField("id")
        if not isinstance(self.id, InstanceNodeId):
            self.id = InstanceNodeId(self.id)

        super().__post_init__(**kwargs)


@dataclass
class Edge(OntologyElement):
    """
    A relationship between two nodes.
    Currently the only kinds of edges supported in KGCL:

    * A subClassOf B <==> Edge(subject=A, predicate=owl:subClassOf, object=B)
    * A subClassOf P some B <==> Edge(subject=A, predicate=P, object=B)
    * P subPropertyOf Q <==> Edge(subject=P, predicate=owl:subPropertyOf, object=Q)

    These represent the most common kind of pairwise relationship between classes, and classes are the dominant node
    type in ontologies.

    In future a wider variety of OWL axiom types will be supportedn through the use of an additional edge
    property/slot to indicate the interpretation of the axiom, following owlstar
    (https://github.com/cmungall/owlstar).
    For example:
    * `A subClassOf R only B <==> Edge(subject=A, predicate=P, object=B, interpretation=AllOnly)`
    * `A Annotation(P,B) <==> Edge(subject=A, predicate=P, object=B, interpretation=annotationAssertion)`

    Note that not all axioms are intended to map to edges. Axioms/triples where the object is a literal would be
    represented as node properties. Complex OWL axioms involving nesting would have their own dedicated construct, or
    may be represented generically. These are out of scope for the current version of KGCL
    """

    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OM.Edge
    class_class_curie: ClassVar[str] = "om:Edge"
    class_name: ClassVar[str] = "edge"
    class_model_uri: ClassVar[URIRef] = OM.Edge

    subject: Optional[Union[str, NodeId]] = None
    predicate: Optional[Union[str, NodeId]] = None
    object: Optional[Union[str, NodeId]] = None
    subject_representation: Optional[str] = None
    predicate_representation: Optional[str] = None
    object_representation: Optional[str] = None
    annotation_set: Optional[Union[dict, Annotation]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self.subject is not None and not isinstance(self.subject, NodeId):
            self.subject = NodeId(self.subject)

        if self.predicate is not None and not isinstance(self.predicate, NodeId):
            self.predicate = NodeId(self.predicate)

        if self.object is not None and not isinstance(self.object, NodeId):
            self.object = NodeId(self.object)

        if self.subject_representation is not None and not isinstance(
            self.subject_representation, str
        ):
            self.subject_representation = str(self.subject_representation)

        if self.predicate_representation is not None and not isinstance(
            self.predicate_representation, str
        ):
            self.predicate_representation = str(self.predicate_representation)

        if self.object_representation is not None and not isinstance(
            self.object_representation, str
        ):
            self.object_representation = str(self.object_representation)

        if self.annotation_set is not None and not isinstance(
            self.annotation_set, Annotation
        ):
            self.annotation_set = Annotation(**as_dict(self.annotation_set))

        super().__post_init__(**kwargs)


class LogicalDefinition(OntologyElement):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OM.LogicalDefinition
    class_class_curie: ClassVar[str] = "om:LogicalDefinition"
    class_name: ClassVar[str] = "logical definition"
    class_model_uri: ClassVar[URIRef] = OM.LogicalDefinition


class OntologySubset(OntologyElement):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = OM.OntologySubset
    class_class_curie: ClassVar[str] = "om:OntologySubset"
    class_name: ClassVar[str] = "ontology subset"
    class_model_uri: ClassVar[URIRef] = OM.OntologySubset


# Enumerations
class OwlTypeEnum(EnumDefinitionImpl):

    CLASS = PermissibleValue(text="CLASS", meaning=OWL.Class)
    OBJECT_PROPERTY = PermissibleValue(
        text="OBJECT_PROPERTY", meaning=OWL.ObjectProperty
    )
    NAMED_INDIVIDUAL = PermissibleValue(
        text="NAMED_INDIVIDUAL", meaning=OWL.NamedIndividual
    )

    _defn = EnumDefinition(
        name="OwlTypeEnum",
    )


class SynonymScopeEnum(EnumDefinitionImpl):

    related = PermissibleValue(text="related", meaning=OIO.hasNarrowSynonym)
    broad = PermissibleValue(text="broad", meaning=OIO.hasBroadSynonym)
    narrow = PermissibleValue(text="narrow", meaning=OIO.hasNarrowSynonym)
    exact = PermissibleValue(text="exact", meaning=OIO.hasExactSynonym)

    _defn = EnumDefinition(
        name="SynonymScopeEnum",
    )


# Slots
class slots:
    pass


slots.owl_type = Slot(
    uri=OM.owl_type,
    name="owl type",
    curie=OM.curie("owl_type"),
    model_uri=OM.owl_type,
    domain=None,
    range=Optional[Union[str, "OwlTypeEnum"]],
)

slots.name = Slot(
    uri=OM.name,
    name="name",
    curie=OM.curie("name"),
    model_uri=OM.name,
    domain=None,
    range=Optional[str],
)

slots.subject = Slot(
    uri=OM.subject,
    name="subject",
    curie=OM.curie("subject"),
    model_uri=OM.subject,
    domain=None,
    range=Optional[Union[str, NodeId]],
)

slots.object = Slot(
    uri=OM.object,
    name="object",
    curie=OM.curie("object"),
    model_uri=OM.object,
    domain=None,
    range=Optional[Union[str, NodeId]],
)

slots.predicate = Slot(
    uri=OM.predicate,
    name="predicate",
    curie=OM.curie("predicate"),
    model_uri=OM.predicate,
    domain=None,
    range=Optional[Union[str, NodeId]],
)

slots.annotation_set = Slot(
    uri=OM.annotation_set,
    name="annotation set",
    curie=OM.curie("annotation_set"),
    model_uri=OM.annotation_set,
    domain=None,
    range=Optional[Union[dict, Annotation]],
)

slots.property = Slot(
    uri=OM.property,
    name="property",
    curie=OM.curie("property"),
    model_uri=OM.property,
    domain=None,
    range=Optional[Union[str, NodeId]],
)

slots.filler = Slot(
    uri=OM.filler,
    name="filler",
    curie=OM.curie("filler"),
    model_uri=OM.filler,
    domain=None,
    range=Optional[str],
)

slots.property_type = Slot(
    uri=OM.property_type,
    name="property type",
    curie=OM.curie("property_type"),
    model_uri=OM.property_type,
    domain=None,
    range=Optional[str],
)

slots.filler_type = Slot(
    uri=OM.filler_type,
    name="filler type",
    curie=OM.curie("filler_type"),
    model_uri=OM.filler_type,
    domain=None,
    range=Optional[str],
)

slots.subject_representation = Slot(
    uri=OM.subject_representation,
    name="subject representation",
    curie=OM.curie("subject_representation"),
    model_uri=OM.subject_representation,
    domain=None,
    range=Optional[str],
)

slots.predicate_representation = Slot(
    uri=OM.predicate_representation,
    name="predicate representation",
    curie=OM.curie("predicate_representation"),
    model_uri=OM.predicate_representation,
    domain=None,
    range=Optional[str],
)

slots.object_representation = Slot(
    uri=OM.object_representation,
    name="object representation",
    curie=OM.curie("object_representation"),
    model_uri=OM.object_representation,
    domain=None,
    range=Optional[str],
)

slots.property_value_set = Slot(
    uri=OM.property_value_set,
    name="property value set",
    curie=OM.curie("property_value_set"),
    model_uri=OM.property_value_set,
    domain=None,
    range=Optional[Union[Union[dict, PropertyValue], List[Union[dict, PropertyValue]]]],
)

slots.id = Slot(
    uri=BASICS.id,
    name="id",
    curie=BASICS.curie("id"),
    model_uri=OM.id,
    domain=None,
    range=URIRef,
)

slots.description = Slot(
    uri=DCTERMS.description,
    name="description",
    curie=DCTERMS.curie("description"),
    model_uri=OM.description,
    domain=None,
    range=Optional[str],
)



"""KGCL parser."""
import logging
import re
import sys
from pathlib import Path
from typing import List

import click
from bioregistry import parse_iri, get_preferred_prefix, curie_to_str
from kgcl_schema.datamodel.kgcl import (
    Change,
    ClassCreation,
    EdgeCreation,
    EdgeDeletion,
    NewSynonym,
    NodeAnnotationChange,
    NodeCreation,
    NodeDeepening,
    NodeDeletion,
    NodeMove,
    NodeObsoletion,
    NodeObsoletionWithDirectReplacement,
    NodeRename,
    NodeShallowing,
    NodeUnobsoletion,
    PlaceUnder,
    PredicateChange,
    RemoveSynonym,
    RemovedNodeFromSubset,
    RemoveUnder,
    Session,
)
from kgcl_schema.datamodel.ontology_model import Edge
from kgcl_schema.utils import to_json, to_rdf, to_yaml
from lark import Lark, Token


def id_generator():
    """Return a new ID for KGCL change operations."""
    id = 0
    while True:
        yield id
        id += 1


# initialise ID generator
id_gen = id_generator()


# initialise parser
path = Path(__file__).parent
kgcl_parser = Lark.open(str(path) + "/kgcl.lark", start="expression")


def parse(input: str) -> List[Change]:
    """
    Parse a set of KGCL command separated by next-line operator.

    Returns instantiated dataclass objects from model.kgcl_schema.
    """
    statements = input.splitlines()
    parsed = []

    for s in statements:
        parsed.append(parse_statement(s))
    return parsed


def parse_statement(input: str) -> Change:
    """
    Parse a KGCL command.

    Return an instantiated dataclass object from model.kgcl_schema.
    """
    # regex = r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+'
    # uri_list = re.findall(regex, input)
    # if uri_list:
    #     # curie = curie_from_iri(uri[0].replace("<", "").replace(">",""))
    #     for _, uri in enumerate(uri_list):
    #         pref, i = parse_iri(uri)
    #         pref = get_preferred_prefix(pref)
    #         curie = curie_to_str(pref, i)
    #         input = input.replace(uri, curie)

    tree = kgcl_parser.parse(input)
    id = "kgcl_change_id_" + str(next(id_gen))

    command = tree.data

    if command == "rename":
        return parse_rename(tree, id)
    elif command == "obsolete":
        return parse_obsolete(tree, id)
    elif command == "unobsolete":
        return parse_unobsolete(tree, id)
    elif command == "delete":
        return parse_delete(tree, id)
    elif command == "move":
        return parse_move(tree, id)
    elif command == "deepen":
        return parse_deepen(tree, id)
    elif command == "shallow":
        return parse_shallow(tree, id)
    elif command == "create_edge":
        return parse_create_edge(tree, id)
    elif command == "delete_edge":
        return parse_delete_edge(tree, id)
    elif command == "change_relationship":
        return parse_change_relationship(tree, id)
    elif command == "change_annotation":
        return parse_change_annotation(tree, id)
    elif command == "create":
        return parse_create(tree, id)
    elif command == "create_class":
        return parse_create_class(tree, id)
    elif command == "create_synonym":
        return parse_create_synonym(tree, id)
    elif command == "remove_from_subset":
        return parse_remove_from_subset(tree, id)
    elif command == "remove_synonym":
        return parse_remove_synonym(tree, id)
    else:
        raise NotImplementedError("No implementation for KGCL command: " + command)

    # TODO: does not have a field for subsets
    # if(command == "add_to_subset"):

    #    term_id = next(tree.find_data('id'))
    #    term_id_token = next(get_tokens(term_id))

    #    subset_id = next(tree.find_data('subset'))
    #    subset_id_token = next(get_tokens(subset_id))

    #    return AddNodeToSubset(
    #        id=id, in_subset=subset_id_token, about_node=term_id_token
    #     )

    # TODO: more commands
    # if(command == "merge"):
    # if(command == "add"):
    # if(command == "add_parent"):
    # if(command == "add_annotation"):
    # if(command == "add_axiom"): (in Manchester Syntax?)
    # if(command == "add_class"):
    # if(command == "add_text_definition"):
    # if(command == "merge"):


def parse_remove_from_subset(tree, id):
    """Remove node from subset."""
    term_id_token = extract(tree, "id")
    subset_id_token = extract(tree, "subset")

    return RemovedNodeFromSubset(
        id=id, subset=subset_id_token, about_node=term_id_token
    )


def parse_create_synonym(tree, id):
    """Create new synonym."""
    entity_token = extract(tree, "entity")
    entity, representation = get_entity_representation(entity_token)

    synonym_string_token = extract(tree, "synonym")
    synonym, representation2 = get_entity_representation(synonym_string_token)

    language_token = extract(tree, "language")
    qualifier_token = extract(tree, "synonym_qualifier")

    return NewSynonym(
        id=id,
        about_node=entity,
        about_node_representation=representation,
        new_value=synonym,
        qualifier=qualifier_token,
        language=language_token,
    )


def parse_remove_synonym(tree, id):
    """Remove a synonym."""
    entity_token = extract(tree, "entity")
    entity, representation = get_entity_representation(entity_token)

    synonym_string_token = extract(tree, "synonym")
    synonym, _ = get_entity_representation(synonym_string_token)

    language_token = extract(tree, "language")
    # qualifier_token = extract(tree, "synonym_qualifier")

    return RemoveSynonym(
        id=id,
        old_value=synonym,
        about_node=entity,
        about_node_representation=representation,
        # qualifier=qualifier_token,
        language=language_token,
    )


def parse_create_class(tree, id):
    """Create new class."""
    term_id_token = extract(tree, "id")
    entity, representation = get_entity_representation(term_id_token)

    return ClassCreation(
        id=id, node_id=entity, about_node_representation=representation
    )


# the KGCL model suggests the command
# 'create node {id} {label} with {annotation set}'
# TODO: handling of {annotation set}
def parse_create(tree, id):
    """Create a node."""
    term_id_token = extract(tree, "id")
    label_token = extract(tree, "label")
    language_token = extract(tree, "language")

    entity, representation = get_entity_representation(term_id_token)

    return NodeCreation(
        id=id,
        about_node=entity,
        about_node_representation=representation,
        node_id=entity,  # was term_id_token
        name=label_token,
        language=language_token,
    )


def parse_change_annotation(tree, id):
    """Change node annotation."""
    subject_token = extract(tree, "entity_subject")
    predicate_token = extract(tree, "entity_predicate")

    subject, s_representation = get_entity_representation(subject_token)
    predicate, p_representation = get_entity_representation(predicate_token)

    old_token = extract(tree, "old_entity_object")
    new_token = extract(tree, "new_entity_object")

    old, old_representation = get_entity_representation(old_token)
    new, new_representation = get_entity_representation(new_token)

    old_language_token = extract(tree, "old_language")
    new_language_token = extract(tree, "new_language")

    old_datatype_token = extract(tree, "old_datatype")
    new_datatype_token = extract(tree, "new_datatype")

    return NodeAnnotationChange(
        id=id,
        about_node=subject,
        about_node_representation=s_representation,
        annotation_property=predicate,
        annotation_property_type=p_representation,
        old_value=old,
        new_value=new,
        old_value_type=old_representation,
        new_value_type=new_representation,
        old_language=old_language_token,
        new_language=new_language_token,
        old_datatype=old_datatype_token,
        new_datatype=new_datatype_token,
    )


def parse_change_relationship(tree, id):
    """Change predicate."""
    subject_token = extract(tree, "entity_subject")
    object_token = extract(tree, "entity_object")

    subject, s_representation = get_entity_representation(subject_token)
    object, o_representation = get_entity_representation(object_token)

    old_token = extract(tree, "old_entity")
    new_token = extract(tree, "new_entity")

    old, old_representation = get_entity_representation(old_token)
    new, new_representation = get_entity_representation(new_token)

    edge = Edge(
        subject=subject,
        predicate=old,
        object=object,
        subject_representation=s_representation,
        predicate_representation=old_representation,
        object_representation=o_representation,
    )

    language_token = extract(tree, "language")
    datatype_token = extract(tree, "datatype")

    return PredicateChange(
        id=id,
        about_edge=edge,
        old_value=old,
        new_value=new,
        old_value_type=old_representation,
        new_value_type=new_representation,
        language=language_token,
        datatype=datatype_token,
    )


def parse_delete_edge(tree, id):
    """Delete edge."""
    subject_token = extract(tree, "entity_subject")
    predicate_token = extract(tree, "entity_predicate")
    object_token = extract(tree, "entity_object_id")

    subject, s_representation = get_entity_representation(subject_token)
    predicate, p_representation = get_entity_representation(predicate_token)
    object, o_representation = get_entity_representation(object_token)

    if (
        predicate == "rdfs:subClassOf"
        or predicate == "<http://www.w3.org/2000/01/rdf-schema#subClassOf>"
    ):
        return RemoveUnder(
            id=id,
            subject=subject,
            predicate=predicate,
            object=object,
            subject_type=s_representation,
            predicate_type=p_representation,
            object_type=o_representation,
        )
    else:
        return EdgeDeletion(
            id=id,
            subject=subject,
            predicate=predicate,
            object=object,
            subject_type=s_representation,
            predicate_type=p_representation,
            object_type=o_representation,
        )


def parse_create_edge(tree, id):
    """Create new edge."""
    subject_token = extract(tree, "entity_subject")
    predicate_token = extract(tree, "entity_predicate")
    object_token = extract(tree, "entity_object_id")

    subject, s_representation = get_entity_representation(subject_token)
    predicate, p_representation = get_entity_representation(predicate_token)
    object, o_representation = get_entity_representation(object_token)

    if (
        predicate == "rdfs:subClassOf"
        or predicate == "<http://www.w3.org/2000/01/rdf-schema#subClassOf>"
    ):
        return PlaceUnder(
            id=id,
            subject=subject,
            predicate=predicate,
            object=object,
            subject_type=s_representation,
            predicate_type=p_representation,
            object_type=o_representation,
        )
    else:
        return EdgeCreation(
            id=id,
            subject=subject,
            predicate=predicate,
            object=object,
            subject_type=s_representation,
            predicate_type=p_representation,
            object_type=o_representation,
        )


def parse_shallow(tree, id):
    """Perform node shallowing."""
    entity_token = extract(tree, "entity")
    old_token = extract(tree, "old_entity")
    new_token = extract(tree, "new_entity")

    entity, e_representation = get_entity_representation(entity_token)
    old_entity, o_representation = get_entity_representation(old_token)
    new_entity, n_representation = get_entity_representation(new_token)

    edge = Edge(
        subject=entity,
        object=old_entity,
        subject_representation=e_representation,
        object_representation=o_representation,
    )

    return NodeShallowing(
        id=id,
        about_edge=edge,
        old_value=old_entity,
        new_value=new_entity,
        old_object_type=o_representation,
        new_object_type=n_representation,
    )


def parse_deepen(tree, id):
    """Perform node deepening."""
    entity_token = extract(tree, "entity")
    old_token = extract(tree, "old_entity")
    new_token = extract(tree, "new_entity")

    entity, e_representation = get_entity_representation(entity_token)
    old_entity, o_representation = get_entity_representation(old_token)
    new_entity, n_representation = get_entity_representation(new_token)

    edge = Edge(
        subject=entity,
        object=old_entity,
        subject_representation=e_representation,
        object_representation=o_representation,
    )

    return NodeDeepening(
        id=id,
        about_edge=edge,
        old_value=old_entity,
        new_value=new_entity,
        old_object_type=o_representation,
        new_object_type=n_representation,
    )


def parse_move(tree, id):
    """Move a node."""
    subject_token = extract(tree, "entity_subject")
    predicate_token = extract(tree, "entity_predicate")
    object_token = extract(tree, "entity_object")

    subject, s_representation = get_entity_representation(subject_token)
    predicate, p_representation = get_entity_representation(predicate_token)
    object, o_representation = get_entity_representation(object_token)

    old_token = extract(tree, "old_entity")
    new_token = extract(tree, "new_entity")

    old, old_representation = get_entity_representation(old_token)
    new, new_representation = get_entity_representation(new_token)

    edge = Edge(
        subject=subject,
        predicate=predicate,
        object=object,
        subject_representation=s_representation,
        predicate_representation=p_representation,
        object_representation=o_representation,
    )

    return NodeMove(
        id=id,
        about_edge=edge,
        old_value=old,
        new_value=new,
        old_object_type=old_representation,
        new_object_type=new_representation,
    )


def parse_delete(tree, id):
    """Delete a node."""
    entity_token = extract(tree, "entity")
    entity, representation = get_entity_representation(entity_token)

    return NodeDeletion(
        id=id, about_node=entity, about_node_representation=representation
    )


def parse_unobsolete(tree, id):
    """Unobsolete a node."""
    entity_token = extract(tree, "entity")
    entity, representation = get_entity_representation(entity_token)

    return NodeUnobsoletion(
        id=id, about_node=entity, about_node_representation=representation
    )


def parse_obsolete(tree, id):
    """Obsolete a node."""
    entity_token = extract(tree, "entity")
    entity, representation = get_entity_representation(entity_token)

    replacement_token = extract(tree, "replacement")
    if replacement_token:
        return NodeObsoletionWithDirectReplacement(
            id=id,
            about_node=entity,
            about_node_representation=representation,
            has_direct_replacement=replacement_token,
        )
    else:
        return NodeObsoletion(
            id=id, about_node=entity, about_node_representation=representation
        )


def parse_rename(tree, id):
    """Rename a node."""
    old_token = extract(tree, "old_label")
    new_token = extract(tree, "new_label")
    old_language = extract(tree, "old_language")
    new_language = extract(tree, "new_language")

    # TODO old_token and new_token are enclosed in ''

    term_id_token = extract(tree, "id")
    if term_id_token is not None:
        entity, representation = get_entity_representation(term_id_token)
    else:
        entity = None
        representation = None

    return NodeRename(
        id=id,
        about_node=entity,
        about_node_representation=representation,
        old_value=old_token,
        new_value=new_token,
        old_language=old_language,
        new_language=new_language,
    )


def extract(tree, data):
    """Extract node."""
    node = get_next(tree.find_data(data))
    if node is not None:
        node_token = next(get_tokens(node))
        return node_token
    else:
        return None


def get_tokens(tree):
    """Get tokens."""
    return tree.scan_values(lambda v: isinstance(v, Token))


def get_next(generator):
    """Next generator."""
    try:
        res = next(generator)
        return res
    except StopIteration:
        return None


def get_entity_representation(entity):
    """Get entity representation."""
    first_character = entity[0]
    last_character = entity[-1:]
    if first_character == "<" and last_character == ">":
        return (
            contract_uri(entity.replace("<", "").replace(">", "")),
            "curie",
        )  # removing brackets
    if first_character == "'" and last_character == "'" and entity[1] != "'":
        return entity[1:-1], "label"
    if first_character == '"' and last_character == '"':
        return entity[1:-1], "literal"
    if entity[0:3] == '"""' and entity[-3:] == '"""':
        return entity[3:-3], "literal"
    if entity[0:3] == "'''" and entity[-3:] == "'''":
        return entity[3:-3], "literal"

    # TODO: use predefined set of prefixes to identify CURIEs
    return contract_uri(str(entity)), "curie"
    # return entity, "error"


def contract_uri(uri_or_curie: str):
    if uri_or_curie.startswith("http://") or uri_or_curie.startswith("https://"):
        pref, i = parse_iri(uri_or_curie)
        pref = get_preferred_prefix(pref)
        curie = curie_to_str(pref, i)
        return curie
    else:
        return uri_or_curie


@click.command()
@click.option("--output", "-o", type=click.File(mode="w"), default=sys.stdout)
@click.option("--output-type", "-O", help="format for output")
@click.option("-v", "--verbose", count=True)
@click.argument("patches", nargs=-1)
def cli(patches, verbose: int, output, output_type):
    """
    Parse a patch or patches specified in KGCL DSL
    """
    if verbose >= 2:
        logging.basicConfig(level=logging.DEBUG)
    elif verbose == 1:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)

    # parser kgcl_schema commands
    changes = []
    for patch in patches:
        logging.info(f"Patch: {patch}")
        changes += parse(patch)
    session = Session(change_set=changes)
    if output_type is None or output_type == "yaml":
        output.write(to_yaml(session))
    elif output_type == "json":
        output.write(to_json(session))
    elif output_type == "rdf":
        output.write(to_rdf(session))
    else:
        raise NotImplementedError(output_type)


if __name__ == "__main__":
    cli()



# TODO: Remove this - it is mis-named and doesn't do anything

"""KGCL RDF-related operations."""
from deprecated.classic import deprecated
from kgcl_schema.grammar.parser import parse
from kgcl_schema.datamodel.kgcl import (
    ClassCreation,
    EdgeCreation,
    EdgeDeletion,
    NewSynonym,
    NodeAnnotationChange,
    NodeCreation,
    NodeDeepening,
    NodeDeletion,
    NodeMove,
    NodeObsoletion,
    NodeRename,
    NodeShallowing,
    NodeUnobsoletion,
    PlaceUnder,
    PredicateChange,
    RemoveUnder,
)

# parse kgcl_schema to python data class
# serialse puython data class to yaml
# convert python data class to RDF


@deprecated("Use builtin linkml yaml_dumper")
def kgcl_2_yaml(kgcl_patch, output):
    """
    Generate YAML from KGCL.

    :param kgcl_patch: KGCL
    :param output: Serialization
    """
    parsed_patch = parse(kgcl_patch)
    serialise_yaml(parsed_patch, output)


@deprecated("Use builtin linkml yaml_dumper")
def serialise_yaml(kgcl_instances, output):
    """Write YAML file based on KGCL instances."""
    preamble = """activity_set:
- {id: uuid:876aac6a6ef6, description: "my sane edits", was_associated_with: "https://orcid.org/0000-0002-6601-2165"}
- {id: uuid:e887d3aac33f, description: "deliberate mistakes", was_associated_with: "OntologicalAnarchist"}
change_set:\n"""

    with open(output, "w") as f:
        f.write(preamble)
        for k in kgcl_instances:
            f.write(data_model_to_yaml(k))
            f.write("\n")


@deprecated("Use builtin linkml yaml_dumper")
def data_model_to_yaml(kgcl_instance):
    """
    Convert a data model to YAML.

    :param kgcl_instance: KGCL instance.
    :return: Appropriate function.
    """
    if type(kgcl_instance) is NodeRename:
        return rename(kgcl_instance)
    if type(kgcl_instance) is NodeObsoletion:
        return obsoletion(kgcl_instance)
    if type(kgcl_instance) is NodeUnobsoletion:
        return unobsoletion(kgcl_instance)
    if type(kgcl_instance) is NodeDeletion:
        return node_deletion(kgcl_instance)
    if type(kgcl_instance) is NodeMove:
        return node_move(kgcl_instance)
    if type(kgcl_instance) is NodeDeepening:
        return node_deepening(kgcl_instance)
    if type(kgcl_instance) is NodeShallowing:
        return node_shallowing(kgcl_instance)
    if type(kgcl_instance) is EdgeCreation:
        return edge_creation(kgcl_instance)
    if type(kgcl_instance) is EdgeDeletion:
        return edge_deletion(kgcl_instance)
    if type(kgcl_instance) is PredicateChange:
        return predicate_change(kgcl_instance)
    if type(kgcl_instance) is NodeAnnotationChange:
        return node_annotation_change(kgcl_instance)
    if type(kgcl_instance) is NodeCreation:
        return node_creation(kgcl_instance)
    if type(kgcl_instance) is ClassCreation:
        return class_creation(kgcl_instance)
    if type(kgcl_instance) is NewSynonym:
        return new_synonym(kgcl_instance)
    if type(kgcl_instance) is PlaceUnder:
        return place_under(kgcl_instance)
    if type(kgcl_instance) is RemoveUnder:
        return place_under(kgcl_instance)


def rename(kgcl_instance):
    """Rename serialization."""
    type = "NodeRename"
    id = kgcl_instance.id
    about_node = render_id(
        kgcl_instance.about_node, kgcl_instance.about_node_representation
    )
    old_value = kgcl_instance.old_value
    new_value = kgcl_instance.new_value
    old_language = kgcl_instance.old_language
    new_language = kgcl_instance.new_language

    serialisation = (
        f"type: {type}, "
        f"id: {id}, "
        f"about_node: {about_node}, "
        f"old_value: {old_value}, "
        f"new_value: {new_value}, "
        f"old_language: {old_language}, "
        f"new_language: {new_language}"
    )

    return "- {" + serialisation + "}"


def obsoletion(kgcl_instance):
    """Obsoletion serialization."""
    type = "NodeObsoletion"
    id = kgcl_instance.id
    about_node = render_id(
        kgcl_instance.about_node, kgcl_instance.about_node_representation
    )
    replacement = kgcl_instance.has_direct_replacement

    serialisation = (
        f"type: {type}, "
        f"id: {id}, "
        f"about_node: {about_node}, "
        f"has_direct_replacement: {replacement} "
    )

    return "- {" + serialisation + "}"


def unobsoletion(kgcl_instance):
    """Unobsoletion serialization."""
    type = "NodeUnobsoletion"
    id = kgcl_instance.id
    about_node = render_id(
        kgcl_instance.about_node, kgcl_instance.about_node_representation
    )

    serialisation = f"type: {type}, " f"id: {id}, " f"about_node: {about_node}"

    return "- {" + serialisation + "}"


def node_deletion(kgcl_instance):
    """Node delation serialization."""
    type = "NodeDeletion"
    id = kgcl_instance.id
    about_node = render_id(
        kgcl_instance.about_node, kgcl_instance.about_node_representation
    )

    serialisation = f"type: {type}, " f"id: {id}, " f"about_node: {about_node}"

    return "- {" + serialisation + "}"


def node_move(kgcl_instance):
    """Node move serialization."""
    type = "NodeMove"
    id = kgcl_instance.id
    subject = render_id(
        kgcl_instance.about_edge.subject,
        kgcl_instance.about_edge.subject_representation,
    )
    predicate = render_id(
        kgcl_instance.about_edge.predicate,
        kgcl_instance.about_edge.predicate_representation,
    )
    object = render_id(
        kgcl_instance.about_edge.object,
        kgcl_instance.about_edge.object_representation,
    )
    about_edge = (
        f"{{"
        f"subject: {subject}, "
        f"predicate: {predicate}, "
        f"object: {object} "
        f"}}"
    )

    new_value = render_id(kgcl_instance.new_value, kgcl_instance.new_object_type)
    old_value = render_id(kgcl_instance.old_value, kgcl_instance.old_object_type)

    serialisation = (
        f"type: {type}, "
        f"id: {id}, "
        f"about_edge: {about_edge},"
        f"new_value: {new_value}, "
        f"old_value: {old_value} "
    )

    return "- {" + serialisation + "}"


def node_deepening(kgcl_instance):
    """Node deopening serialization."""
    type = "NodeDeepening"
    id = kgcl_instance.id
    subject = render_id(
        kgcl_instance.about_edge.subject,
        kgcl_instance.about_edge.subject_representation,
    )
    predicate = render_id(
        kgcl_instance.about_edge.predicate,
        kgcl_instance.about_edge.predicate_representation,
    )
    object = render_id(
        kgcl_instance.about_edge.object,
        kgcl_instance.about_edge.object_representation,
    )
    about_edge = (
        f"{{"
        f"subject: {subject}, "
        f"predicate: {predicate}, "
        f"object: {object} "
        f"}}"
    )

    new_value = render_id(kgcl_instance.new_value, kgcl_instance.new_object_type)
    old_value = render_id(kgcl_instance.old_value, kgcl_instance.old_object_type)

    serialisation = (
        f"type: {type}, "
        f"id: {id}, "
        f"about_edge: {about_edge},"
        f"new_value: {new_value}, "
        f"old_value: {old_value} "
    )

    return "- {" + serialisation + "}"


def node_shallowing(kgcl_instance):
    """Node shallowing serialization."""
    type = "NodeShallowing"
    id = kgcl_instance.id
    subject = render_id(
        kgcl_instance.about_edge.subject,
        kgcl_instance.about_edge.subject_representation,
    )
    predicate = render_id(
        kgcl_instance.about_edge.predicate,
        kgcl_instance.about_edge.predicate_representation,
    )
    object = render_id(
        kgcl_instance.about_edge.object,
        kgcl_instance.about_edge.object_representation,
    )
    about_edge = (
        f"{{"
        f"subject: {subject}, "
        f"predicate: {predicate}, "
        f"object: {object} "
        f"}}"
    )

    new_value = render_id(kgcl_instance.new_value, kgcl_instance.new_object_type)
    old_value = render_id(kgcl_instance.old_value, kgcl_instance.old_object_type)

    serialisation = (
        f"type: {type}, "
        f"id: {id}, "
        f"about_edge: {about_edge},"
        f"new_value: {new_value}, "
        f"old_value: {old_value} "
    )

    return "- {" + serialisation + "}"


def edge_creation(kgcl_instance):
    """Edge creation serialization."""
    type = "EdgeCreation"
    id = kgcl_instance.id
    subject = render_id(
        kgcl_instance.subject,
        kgcl_instance.subject_type,
    )
    predicate = render_id(
        kgcl_instance.predicate,
        kgcl_instance.predicate_type,
    )
    object = render_id(
        kgcl_instance.object,
        kgcl_instance.object_type,
    )

    serialisation = (
        f"type: {type}, "
        f"id: {id}, "
        f"subject: {subject}, "
        f"predicate: {predicate}, "
        f"object: {object} "
    )

    return "- {" + serialisation + "}"


def edge_deletion(kgcl_instance):
    """Edge deletion serialization."""
    type = "EdgeDeletion"
    id = kgcl_instance.id
    subject = render_id(
        kgcl_instance.subject,
        kgcl_instance.subject_type,
    )
    predicate = render_id(
        kgcl_instance.predicate,
        kgcl_instance.predicate_type,
    )
    object = render_id(
        kgcl_instance.object,
        kgcl_instance.object_type,
    )

    serialisation = (
        f"type: {type}, "
        f"id: {id}, "
        f"subject: {subject}, "
        f"predicate: {predicate}, "
        f"object: {object} "
    )

    return "- {" + serialisation + "}"


def predicate_change(kgcl_instance):
    """Predicate change serialization."""
    type = "PredicateChange"
    id = kgcl_instance.id
    subject = render_id(
        kgcl_instance.about_edge.subject,
        kgcl_instance.about_edge.subject_representation,
    )
    predicate = render_id(
        kgcl_instance.about_edge.predicate,
        kgcl_instance.about_edge.predicate_representation,
    )
    object = render_id(
        kgcl_instance.about_edge.object,
        kgcl_instance.about_edge.object_representation,
    )
    about_edge = (
        f"{{"
        f"subject: {subject}, "
        f"predicate: {predicate}, "
        f"object: {object} "
        f"}}"
    )
    old_value = kgcl_instance.old_value
    new_value = kgcl_instance.new_value
    old_language = kgcl_instance.old_language
    new_language = kgcl_instance.new_language

    serialisation = (
        f"type: {type}, "
        f"id: {id}, "
        f"about_edge: {about_edge},"
        f"old_value: {old_value}, "
        f"new_value: {new_value}, "
        f"old_language: {old_language}, "
        f"new_language: {new_language}"
    )

    return "- {" + serialisation + "}"


def node_annotation_change(kgcl_instance):
    """Node annotation serialization."""
    type = "NodeAnnotationChange"
    id = kgcl_instance.id
    about_node = render_id(
        kgcl_instance.about_node, kgcl_instance.about_node_representation
    )
    old_value = kgcl_instance.old_value
    new_value = kgcl_instance.new_value
    old_language = kgcl_instance.old_language
    new_language = kgcl_instance.new_language
    old_datatype = kgcl_instance.old_datatype
    new_datatype = kgcl_instance.new_datatype

    serialisation = (
        f"type: {type}, "
        f"id: {id}, "
        f"about_node: {about_node},"
        f"old_value: {old_value}, "
        f"new_value: {new_value}, "
        f"old_language: {old_language}, "
        f"new_language: {new_language},"
        f"old_datatype: {old_datatype}, "
        f"new_datatype: {new_datatype}"
    )

    return "- {" + serialisation + "}"


def node_creation(kgcl_instance):
    """Node creation serialization."""
    type = "NodeCreation"
    id = kgcl_instance.id
    about_node = render_id(
        kgcl_instance.about_node, kgcl_instance.about_node_representation
    )
    name = kgcl_instance.name
    language = kgcl_instance.language

    serialisation = (
        f"type: {type}, "
        f"id: {id}, "
        f"about_node: {about_node}, "
        f"language: {language},"
        f"name: {name}"
    )

    return "- {" + serialisation + "}"


def class_creation(kgcl_instance):
    """Class creation serialization."""
    type = "ClassCreation"
    id = kgcl_instance.id
    about_node = render_id(
        kgcl_instance.about_node, kgcl_instance.about_node_representation
    )

    serialisation = f"type: {type}, " f"id: {id}, " f"about_node: {about_node}, "

    return "- {" + serialisation + "}"


def new_synonym(kgcl_instance):
    """Return serialization for new synonym."""
    type = "NewSynonym"
    id = kgcl_instance.id
    about_node = render_id(
        kgcl_instance.about_node, kgcl_instance.about_node_representation
    )
    new_value = kgcl_instance.new_value
    qualifier = kgcl_instance.qualifier
    language = kgcl_instance.language

    serialisation = (
        f"type: {type}, "
        f"id: {id}, "
        f"about_node: {about_node}, "
        f"language: {language},"
        f"new_value: {new_value},"
        f"qualifier: {qualifier}"
    )

    return "- {" + serialisation + "}"


def place_under(kgcl_instance):
    """Place under serialization."""
    type = "PlaceUnder"
    id = kgcl_instance.id
    subject = render_id(
        kgcl_instance.subject,
        kgcl_instance.subject_type,
    )
    predicate = render_id(
        kgcl_instance.predicate,
        kgcl_instance.predicate_type,
    )
    object = render_id(
        kgcl_instance.object,
        kgcl_instance.object_type,
    )

    serialisation = (
        f"type: {type}, "
        f"id: {id}, "
        f"subject: {subject}, "
        f"predicate: {predicate}, "
        f"object: {object} "
    )

    return "- {" + serialisation + "}"


def remove_under(kgcl_instance):
    """Remove under serialization."""
    type = "RemoveUnder"
    id = kgcl_instance.id
    subject = render_id(
        kgcl_instance.subject,
        kgcl_instance.subject_type,
    )
    predicate = render_id(
        kgcl_instance.predicate,
        kgcl_instance.predicate_type,
    )
    object = render_id(
        kgcl_instance.object,
        kgcl_instance.object_type,
    )

    serialisation = (
        f"type: {type}, "
        f"id: {id}, "
        f"subject: {subject}, "
        f"predicate: {predicate}, "
        f"object: {object} "
    )

    return "- {" + serialisation + "}"


def render_id(id, type):
    """Render id."""
    if type == "label":
        return '"' + str(id) + '"'
    else:
        return str(id)



"""Convenience wrappers around linkml runtime.

Needed for doing basic conversion between objects and serialization formats.
The top level class is a Session object
Some of this will become unnecessary in the future
"""

import json
import logging
import os
from typing import List, Dict, Any, Union
from uuid import uuid1

import click
import yaml
from kgcl_schema.schema import get_schemaview
from linkml_runtime.dumpers import json_dumper, yaml_dumper, rdflib_dumper
from linkml_runtime.dumpers.json_dumper import JSONDumper
from linkml_runtime.dumpers.rdf_dumper import RDFDumper
from linkml_runtime.loaders.yaml_loader import YAMLLoader
from linkml_runtime.utils.formatutils import camelcase
from linkml_runtime.utils.yamlutils import YAMLRoot
from rdflib import Graph

import kgcl_schema.datamodel as model
from kgcl_schema.datamodel.kgcl import Session, Change
from kgcl_schema.datamodel.kgcl import Activity
from os.path import dirname, join

# THIS_DIR = os.path.abspath(os.path.dirname(__file__))
# LD = os.path.join(THIS_DIR, "../ldcontext/kgcl_schema.context.jsonld")
LD = join(dirname(dirname(dirname(__file__))), "ldcontext/kgcl.context.jsonld")


def get_context() -> str:
    """Get JSON-LD context."""
    with open(LD) as stream:
        context = json.load(stream)
        context["@context"]["ANAT"] = {
            "@id": "https://example.org/anatomy/",
            "@prefix": True,
        }
        context["@context"]["uuid"] = {
            "@id": "https://example.org/uuid/",
            "@prefix": True,
        }
        context["@context"]["type"] = {"@type": "@id"}
        return json.dumps(context)


def to_json(session: Session) -> str:
    """Convert a session object to plain json."""
    assign_types(session)
    return json_dumper.dumps(session)


def to_yaml(session: Session) -> str:
    """Convert a session object to plain json."""
    assign_types(session)
    return yaml_dumper.dumps(session)


def to_jsonld(session: Session) -> str:
    """Convert a session object to JSON-LD."""
    assign_types(session)
    return json_dumper.dumps(session, get_context())


# TODO: allow passing of prefix maps
def to_rdf(session: Session, prefix_map: Dict[str, str] = None) -> str:
    """Convert a session object to an rdflib Graph string."""
    assign_types(session)
    return rdflib_dumper.dumps(
        session, prefix_map=prefix_map, schemaview=get_schemaview()
    )


def from_json(filename: str) -> Session:
    """
    As from_yaml, but with json
    """
    logging.info(f"Converting {filename}")
    with open(filename, "r") as stream:
        obj = json.load(stream)
        return from_dict(obj)


def from_yaml(filename: str) -> Session:
    """
    Parse a slight variant of the standard serialization of the model into YAML.

    See de-novo.yaml as an example

    In particular, until linkml supports a polymorphic discriminator we need ad-hoc
    code for instantiating to the correct class
    """
    logging.info(f"Converting {filename}")
    with open(filename, "r") as stream:
        obj = yaml.safe_load(stream)
        return from_dict(obj)


def from_dict(obj: Dict[str, Any]) -> Session:
    # TODO: move this to LinkML runtime
    # at this time we need wrapper code to introspect type fields
    # and use this to determine which type to instantiate;
    # we also auto-populate identifier fields
    session = Session()
    loader = YAMLLoader()
    for a in obj.get("activity_set", []):
        session.activity_set.append(Activity(**a))
    for c in obj.get("change_set", []):
        # auto-create an ID
        c["id"] = f"uuid:{uuid1()}"
        tc = c["type"]
        del c["type"]
        typ = getattr(model.kgcl, tc)
        logging.info(f"Converting type {tc} // {c}")
        chg = typ(**c)
        chg.type = f"kgcl_schema:{tc}"
        session.change_set.append(chg)
    return session


def assign_types(obj: YAMLRoot):
    if isinstance(obj, Session):
        for ch in obj.change_set:
            assign_types(ch)
    elif isinstance(obj, Change):
        tc = type(obj).class_name
        obj.type = camelcase(tc)


@click.command()
@click.argument("files", nargs=-1)
def cli(files: List[str]):
    """CLI."""
    pass
    # for f in files:
    #     session = from_yaml(f)


# USED FOR TESTING:
# https://github.com/cmungall/knowledge-graph-change-language/pull/30#discussion_r832205466


if __name__ == "__main__":
    cli()



from pathlib import Path
from linkml_runtime import SchemaView

PATH = Path(__file__).parent


def get_schemaview() -> SchemaView:
    schema_root = str(PATH / "kgcl.yaml")
    return SchemaView(schema_root)
